# Generated by Django 3.1.6 on 2021-02-15 04:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("securepasswords", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="passwordprofile",
            name="force_change",
            field=models.BooleanField(default=False, verbose_name="force change"),
        ),
    ]



# Generated by Django 3.1.6 on 2021-02-08 07:38

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="PasswordProfile",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE, related_name="+", to=settings.AUTH_USER_MODEL
                    ),
                ),
            ],
            options={
                "verbose_name": "password profile",
                "verbose_name_plural": "password profiles",
            },
        ),
        migrations.CreateModel(
            name="Password",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("encoded", models.CharField(editable=False, max_length=128, verbose_name="password")),
                ("created", models.DateTimeField(auto_now_add=True)),
                (
                    "profile",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="password_history",
                        to="securepasswords.passwordprofile",
                    ),
                ),
            ],
            options={
                "verbose_name": "password",
                "verbose_name_plural": "passwords",
                "ordering": ["profile", "-created"],
            },
        ),
    ]



from django.core.management.base import BaseCommand

from securepasswords.models import PasswordProfile


class Command(BaseCommand):
    help = "Enforces password renewal"

    def handle(self, *args, **options):
        PasswordProfile.objects.update(force_change=True)



from typing import Optional

from django.contrib.auth import get_user_model
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver

from securepasswords.models import PasswordProfile
from securepasswords.utils import get_password_profile

__all__ = [
    "password_history_update",
]


@receiver(post_save, sender=get_user_model())
def password_history_update(sender, **kwargs):
    profile = _password_changed(kwargs)
    if profile:
        profile.password_history.create(encoded=profile.user.password)
        profile.force_change = False
        profile.save(update_fields=("force_change",))


def _password_changed(signal_kwargs) -> Optional[PasswordProfile]:
    update_fields = signal_kwargs.get("update_fields", [])
    if not update_fields or "password" in update_fields:
        user = signal_kwargs["instance"]
        profile = get_password_profile(user)
        latest_password = profile.password_history.latest("created")
        if not latest_password or latest_password.encoded != user.password:
            return profile



from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import check_password
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _


class PasswordProfile(models.Model):
    # do not clutter the user namespace
    user = models.OneToOneField(get_user_model(), on_delete=models.CASCADE, related_name="+")
    force_change = models.BooleanField(_("force change"), default=False)

    class Meta:
        verbose_name = _("password profile")
        verbose_name_plural = _("password profiles")

    def has_used_password(self, raw_password: str, last_n: int = None):
        history = self.password_history.all()  # ordered by descending date
        if last_n is not None:
            history = history[:last_n]
        return any(check_password(raw_password, password.encoded) for password in history)

    def password_age(self):
        last_password: Password = self.password_history.first()
        if last_password:
            age = (timezone.now() - last_password.created).days
            return age
        return None


class Password(models.Model):
    profile = models.ForeignKey(
        PasswordProfile, on_delete=models.CASCADE, related_name="password_history", editable=False
    )
    encoded = models.CharField(_("password"), max_length=128, editable=False)
    created = models.DateTimeField(auto_now_add=True, editable=False)

    class Meta:
        verbose_name = _("password")
        verbose_name_plural = _("passwords")
        ordering = ["profile", "-created"]

    def __str__(self):
        return f"{self.profile.user}: {self.created}"



from django.contrib import messages
from django.shortcuts import redirect
from django.urls import NoReverseMatch, reverse
from django.utils.translation import ugettext as _

from .conf import conf
from .utils import get_password_profile


class SecurePasswordMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if request.user.is_authenticated:
            current_path = request.get_full_path()
            try:
                allowed_path = reverse(conf.CHANGE_PASSWORD_URL)
            except NoReverseMatch:
                allowed_path = conf.CHANGE_PASSWORD_URL
            if current_path != allowed_path:
                profile = get_password_profile(request.user)
                if profile.force_change:
                    messages.info(request, _("You must change your password for security reasons!"))
                    return redirect(conf.CHANGE_PASSWORD_URL)

                if conf.MAX_PASSWORD_AGE is not None:
                    age = profile.password_age()
                    if age is not None and age > conf.MAX_PASSWORD_AGE:

                        messages.info(
                            request,
                            _(
                                "Your current password is older than {max_age} days. "
                                "Please change your password now."
                            ).format(max_age=conf.MAX_PASSWORD_AGE),
                        )
                        return redirect(conf.CHANGE_PASSWORD_URL)
        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        return response



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class SecurePasswordsConfig(AppConfig):
    name = "securepasswords"
    verbose_name = _("secure passwords")

    def ready(self):
        super(SecurePasswordsConfig, self).ready()
        from .signal_handlers import password_history_update  # noqa



from itertools import groupby
from typing import List

from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.utils.translation import gettext as _

from securepasswords.conf import conf
from securepasswords.utils import get_password_profile


class HistoryValidator:
    def __init__(self, last_n=None):
        self.last_n = last_n or conf.PASSWORD_HISTORY_LENGTH

    def validate(self, raw_password, user=None):
        if not (user and user.pk):
            return
        password_profile = get_password_profile(user=user)
        if password_profile.has_used_password(raw_password=raw_password, last_n=self.last_n):
            raise ValidationError(
                _("You have used this password before."),
                code="password_reused",
            )

    def get_help_text(self):
        return _("You must not reuse a previous password.")


class RepeatedCharValidator:
    def __init__(self, max_repetitions=2):
        self.n = max_repetitions

    def validate(self, raw_password, user=None):
        if any(len(list(g)) > self.n for _k, g in groupby(raw_password)):
            raise ValidationError(
                _("Your password contains more than {n} consecutive equal characters.").format(n=self.n),
                code="password_repeated_characters",
            )

    def get_help_text(self):
        return _("Your password must not contain more than {n} consecutive equal characters.").format(n=self.n)


class SequenceValidator:
    def __init__(self, max_length=2):
        self.n = max_length

    def _filter(self, char: str):
        return NotImplemented

    def _process(self, char: str):
        return NotImplemented

    def _get_error_text(self):
        return NotImplemented

    def validate(self, raw_password, user=None):
        for k, g in groupby(raw_password, key=self._filter):
            if k and self.has_sequence(list(map(self._process, g))):
                raise ValidationError(self._get_error_text())

    def has_sequence(self, seq: List) -> bool:
        check_length = self.n + 1
        for i in range(len(seq) - self.n):
            sub = seq[i : i + check_length]
            if len({a - b for a, b in zip(sub, sub[1:])}) == 1:
                return True
        return False


class ArithmeticSequenceValidator(SequenceValidator):
    def _filter(self, char: str):
        return char.isdigit()

    def _process(self, char: str):
        return int(char)

    def get_help_text(self):
        return _("Your password must not contain an arithmetic sequence ('456', '642', etc.) longer than {n}").format(
            n=self.n
        )

    def _get_error_text(self):
        return _("Your password contains an arithmetic sequence ('456', '642', etc.) longer than {n}").format(n=self.n)


class AlphabeticSequenceValidator(SequenceValidator):
    def __init__(self, max_length=3):
        super(AlphabeticSequenceValidator, self).__init__(max_length=max_length)

    def _filter(self, char: str):
        return char.isalpha()

    def _process(self, char: str):
        return ord(char)

    def get_help_text(self):
        return _("Your password must not contain an alphabetic sequence ('dcba', 'aceg', etc.) longer than {n}").format(
            n=self.n
        )

    def _get_error_text(self):
        return _("Your password contains an alphabetic sequence ('dcba', 'aceg', etc.) longer than {n}").format(
            n=self.n
        )


_char_class_tests = (str.isupper, str.islower, str.isdigit, str.isspace)


def ispunct(char: str):
    return not any(t(char) for t in _char_class_tests)


class CharacterClassValidator:
    TESTS = _char_class_tests + (ispunct,)

    def __init__(self, min_count: int = 2):
        if min_count not in range(2, 6):
            raise ImproperlyConfigured("The only valid 'min_count' arguments are 2, 3, 4, or 5")
        self.min_count = min_count

    def get_help_text(self):
        return _(
            "Your password must contain characters from at least {n} categories out of the following 5: "
            "Uppercase letters, lowercase letters, digits, punctuation, whitespace"
        ).format(n=self.min_count)

    def validate(self, raw_password, user=None):
        count = sum(any(map(test, raw_password)) for test in self.TESTS)
        if count < self.min_count:
            raise ValidationError(
                _(
                    "Your password contains characters from fewer than {n} categories out of the following 5: "
                    "Uppercase letters, lowercase letters, digits, punctuation, whitespace"
                ).format(n=self.min_count)
            )



from typing import Optional

from django.contrib.auth.base_user import AbstractBaseUser
from django.utils import timezone

from securepasswords.models import Password, PasswordProfile


def get_password_profile(user: AbstractBaseUser) -> PasswordProfile:
    profile, created = PasswordProfile.objects.get_or_create(user=user)
    if created and user.has_usable_password():
        profile.password_history.create(encoded=user.password)
    return profile


def get_current_password_age(user: AbstractBaseUser) -> Optional[int]:
    """
    Return the age of the current password in days
    """
    profile = get_password_profile(user)
    last_password: Password = profile.password_history.first()
    if last_password:
        age = (timezone.now() - last_password.created).days
        return age
    return None



# -*- coding: utf-8 -*-
import os

from setuptools import setup

base_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(base_dir, "VERSION")) as f:
    VERSION = f.read()

with open(os.path.join(base_dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

DESCRIPTION = (
    "django-secure-passwords is a package that adds more password validation options as "
    "often required in large corporations"
)


setup(
    name="django-secure-passwords",
    description=DESCRIPTION,
    version=VERSION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "Django>=2.2",
    ],
    python_requires="~=3.6",
    packages=[
        "securepasswords",
        "securepasswords.migrations",
        "securepasswords.management",
        "securepasswords.management.commands",
    ],
    author="Veit Rückert",
    author_email="veit@blueshoe.de",
    include_package_data=True,
    package_data={"securepasswords": ["locale/*/LC_MESSAGES/*.mo"]},
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1" "",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

DEFAULTS = {
    "MAX_PASSWORD_AGE": (None, int),
    "PASSWORD_HISTORY_LENGTH": (None, int),
    "CHANGE_PASSWORD_URL": ("password_change", str),  # override may be view name or URL
    "RESET_PASSWORD_URL": ("password_reset", str),  # override may be view name or URL
}


class SecurePasswordConfig:
    def __init__(self):
        options = self.get_options()
        self.MAX_PASSWORD_AGE = options.get("MAX_PASSWORD_AGE", None)
        self.PASSWORD_HISTORY_LENGTH = options.get("PASSWORD_HISTORY_LENGTH", None)
        self.CHANGE_PASSWORD_URL = options.get("CHANGE_PASSWORD_URL", None)

    def get_options(self):
        options = {k: v[0] for k, v in DEFAULTS.items()}
        options.update(self.get_settings_overrides())
        return options

    def get_settings_overrides(self):
        overrides = getattr(settings, "SECURE_PASSWORDS", {})
        if not isinstance(overrides, dict):
            raise ImproperlyConfigured("'SECURE_PASSWORDS' settings must be a dictionary")
        for k, v in overrides.items():
            if k in DEFAULTS:
                type_ = DEFAULTS[k][1]
                if not isinstance(v, type_):
                    try:
                        overrides[k] = type_(v)
                    except (TypeError, ValueError):
                        raise ImproperlyConfigured(
                            f"SECURE_PASSWORDS setting '{k}' must be of type '{type_}' or omitted"
                        )
        return overrides


conf = SecurePasswordConfig()



from django.contrib import admin
from django.contrib.auth import get_user_model
from django.utils.translation import ugettext_lazy as _

from securepasswords.models import PasswordProfile


class ProfileInline(admin.StackedInline):
    model = PasswordProfile
    extra = 0
    fields = ("force_change",)

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


user_model = get_user_model()
user_admin_class = admin.site._registry[user_model].__class__


class SecurePasswordUserAdmin(user_admin_class):
    actions = ["force_password_change"]

    def get_inlines(self, request, obj):
        inlines = super().get_inlines(request, obj)
        inlines.append(ProfileInline)
        return inlines

    def force_password_change(self, request, queryset):
        PasswordProfile.objects.filter(user__in=queryset).update(force_change=True)

    force_password_change.short_description = _("Force selected users to change their passwords")


admin.site.unregister(user_model)
admin.site.register(user_model, SecurePasswordUserAdmin)
